/*
 * This shader is based on the tutorials 19-21 from OGLdev and the paper 
 * "Unified Distance Formulas for Halfspace Fog" by Eric Lengyel.
 *
 * References:
 * http://ogldev.atspace.co.uk/index.html, accessed on 06/05/2015.
 * Lengyel, Eric. Unified Distance Formulas for Halfspace Fog. Journal of 
 * Graphics Tools. vol.12, no.2, 2007.
 */
#version 420

const int MAX_LIGHTS = 10;

in vec2 uv;
smooth in vec3 worldNormal;	 // Interpolated normal vector in world space
in vec3 worldPosition;  // Vectorposition in worldspace
in vec4 smCoords[4];
in float fog_dot_Vertex;
in float fog_dot_Camera;
in vec3 rho_Camera;


struct BaseLight {
	vec3	color;
	float	ambientIntensity;
	float	diffuseIntensity;
};

struct DirectionalLight {
	BaseLight	base;
	vec3		direction;
};

struct Attenuation {
	float	constant;
	float	linear;
	float	exponential;
};

struct SpotLight {
	BaseLight	base;
	vec3		position;
	vec3		direction;
	Attenuation	falloff;
	float		angle;
};


layout(binding = 0) uniform sampler2D texture;
layout(binding = 1) uniform sampler2DShadow shadowMap1;
layout(binding = 2) uniform sampler2DShadow shadowMap2;
layout(binding = 3) uniform sampler2DShadow shadowMap3;
layout(binding = 4) uniform sampler2DShadow shadowMap4;
uniform DirectionalLight moon;
uniform SpotLight lamps[MAX_LIGHTS];
uniform vec3 worldEye;
uniform float specularIntensity;
uniform float specularPower;
uniform vec3 fogColor;
uniform bool renderFog;
uniform float fogLightDistortion;
uniform float moonShadowIntensity;
uniform float opacity;
uniform int isTransparency;

out vec4 fragmentColor;



float calcShadowFactor(vec4 coords, sampler2DShadow shadowMap, float min)
{
	/*vec2 sm_uvs = coords.xy / coords.w;
	float depth_p = coords.z / coords.w;
	float depth_c = textureProj(shadowMap, coords).r;
	return (depth_c < depth_p) ? min : 1.0;*/
	float depth = textureProj(shadowMap, coords).r;

	depth += min;
	if (depth > 1.0)
		depth = 1.0;

	return depth;
}

vec4 calculateLight(BaseLight light, vec3 lightDirection, vec3 normal, 
	float shadowFactor)
{
	vec4 ambientColor = vec4(light.color, 1.0f) * light.ambientIntensity;
	float diffuseFactor = dot(normal, -lightDirection);

	vec4 diffuseColor = vec4(0, 0, 0, 0);
	vec4 specularColor = vec4(0, 0, 0, 0);

	if(diffuseFactor > 0)
	{
		diffuseColor =  vec4(light.color, 1.0f) * 
						light.diffuseIntensity * 
						diffuseFactor;
		
		vec3 vertexToEye = normalize(worldEye - worldPosition);
		vec3 lightReflect = normalize(reflect(lightDirection, normal));
		float specularFactor = dot(vertexToEye, lightReflect);
		specularFactor = pow(specularFactor, specularPower);
		if(specularFactor > 0) 
		{
			specularColor = vec4(light.color, 1.0f) *
							specularIntensity *
							specularFactor;
		}
	}

	return (ambientColor + shadowFactor * (diffuseColor + specularColor));
}

vec4 calculateSpotLight(SpotLight spot, vec3 normal, vec4 lightSpacePos,
	sampler2DShadow shadowMap, float min)
{
	vec3 lightDirection = worldPosition - spot.position;
	vec3 lightToPixel = normalize(lightDirection);
	float spotFactor = dot(lightToPixel, spot.direction);
	float shadowFactor = calcShadowFactor(lightSpacePos, shadowMap, min);

	if (spotFactor > spot.angle && shadowFactor != 0)
	{
		float distance = length(lightDirection);
		vec4 color = calculateLight(spot.base,
									normalize(lightDirection),
									normal,
									shadowFactor);
		float attenuation = spot.falloff.constant +
							spot.falloff.linear * distance +
							spot.falloff.exponential * distance * distance;
		
		// Color divided by attenuation creates the fading effect.
		// In the second row, we are mapping the interval [angle-1] to the
		// range [0-1] to provide much better results!
		return (color / attenuation) * 
			   (1.0 - (1.0 - spotFactor) / (1.0 - spot.angle));
	} else
		return vec4(0, 0, 0, 0);
}

vec3 calculateFog(vec4 originalColor, vec3 additionalLight)
{
	// Fog under a spotlight is lighter so we add the additional brightness
	// from the spotlights the the color of the fog and we can control the
	// magnitude with the constant fogLightDistortion.
	vec3 additionFogLightness = additionalLight * fogLightDistortion;

	// Calculate distance * rho
	float d = clamp(- fog_dot_Vertex / abs(fog_dot_Camera), 0.0, 1.0);
	d *= length(rho_Camera);

	// Calculate fog fraction and apply
	float fogFactor = clamp(exp2(-d), 0.0, 1.0);

	return mix(fogColor + additionFogLightness, originalColor.rgb, fogFactor);
}

void main()
{
	vec3 normal = normalize(worldNormal);
	float shadowFactorMoon = 
		calcShadowFactor(smCoords[0], shadowMap1, moonShadowIntensity);
	vec4 totalLight = 
		calculateLight(moon.base, moon.direction, normal, shadowFactorMoon);
	
	// Calculate addition light from spotlights
	vec4 additionalLight = vec4(0,0,0,0);
	additionalLight +=
		calculateSpotLight(lamps[0], normal, smCoords[1], shadowMap2, 0.0);
	additionalLight +=
		calculateSpotLight(lamps[1], normal, smCoords[2], shadowMap3, 0.0);
	additionalLight +=
		calculateSpotLight(lamps[2], normal, smCoords[3], shadowMap4, 0.0);
	
	totalLight += additionalLight;
	vec4 color = texture2D(texture, uv) * totalLight;

	// Render fog if needed
	if(renderFog)
		fragmentColor.rgb = calculateFog(color, additionalLight.rgb);
	else
		fragmentColor.rgb = color.rgb;

	// Set opacity if isTransparency
	if (isTransparency != 0)
		fragmentColor.a = opacity;
	else
		fragmentColor.a = 1.0f;
}
